/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the External Editor extension.
 *
 * The Initial Developer of the Original Code is Philip Nilsson.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Kimitake <webmaster@kimitake.blogdns.net>
 *   Stas Fomin <stas-fomin@yandex.ru>
 *
 * ***** END LICENSE BLOCK ***** */

function ExternalEditor()
{
	if (!document.getElementById("bundle_externaleditor"))
		return; // Called from the inheritor >_<
}

ExternalEditor.prototype = {
	_prefBranch : null,
	_locale : null,

	Initialize : function() {
		this.addCheckFocus();
		var prefs = Components.classes["@mozilla.org/preferences-service;1"].
			getService(Components.interfaces.nsIPrefService);
		this._prefBranch = prefs.getBranch("extensions.externaleditor.");

		this._locale = document.getElementById("bundle_externaleditor");
		if (!this._locale) {
			this.reportError("Could not load locale strings.");
		}
	},

	// TODO: Make it report the line where this function was called from.
	reportError : function(errorString) {
		if (Components.utils.reportError) {
			Components.utils.reportError(errorString);
		} else {
			Components.classes['@mozilla.org/consoleservice;1']
				.getService(Components.interfaces.nsIConsoleService)
				.logStringMessage(errorString);
		}
	},

	openOptions : function(reason) {
		var features;
		try {
			var prefs = Components.classes["@mozilla.org/preferences-service;1"]
				.getService(Components.interfaces.nsIPrefBranch);
			var instantApply = prefs.getBoolPref("browser.preferences.instantApply");
			features = "chrome,titlebar,toolbar,centerscreen" + (instantApply ? ",dialog=no" : ",modal");
		} catch (e) {
			features = "chrome,titlebar,toolbar,centerscreen,modal";
		}
		window.openDialog('chrome://externaleditor/content/options.xul', 'externaleditor_prefwindow', features, reason);
	},

	// this != gExternalEditor
	checkFocus : function() {
		if (gContextMenu) {
			gContextMenu.showItem("context-externaleditor-edit",
				gContextMenu.onTextInput);
		}
	},

	// This is only relevant to Firefox.
	// The menu popup is showing all the time in Thunderbird.
	// 2006-01-16 leffe: What did I mean by that?
	addCheckFocus : function() {
		var menu = document.getElementById("contentAreaContextMenu");
		if (menu) {
			menu.addEventListener("popupshowing", this.checkFocus, false);
		}
	},

	// TODO: Supply a way of setting the caret back
	//       after the data has been rewritten.
	getCaretLine : function() {
		var area = gContextMenu.target;
		var data = area.value.substring(0, area.selectionStart);
		var res = data.match(/(\r\n|\r|\n)/g);
		if (res == null)
			return 0;
		// Add 1 for some reason... seems to work :)
		return res.length + 1;
	},

	editFile : function(filename) {
		// Figure out what editor to use.
		var file = Components.classes["@mozilla.org/file/local;1"]
			.createInstance(Components.interfaces.nsILocalFile);
		var argumentList = this._prefBranch.getCharPref("arguments");
		var editor = this._prefBranch.getCharPref("editor");
		if (!editor) {
			this.reportError(this._locale.getString("Error_unset"));
			this.openOptions(this._locale.getString("Error_unset"));
			return false;
		}
		file.initWithPath(editor);
		if (!file.exists()) {
			this.reportError(this._locale.getString("Error_invalid_file"));
			this.openOptions(this._locale.getString("Error_invalid_file"));
			return false;
		}
		if (!file.isExecutable()) {
			this.reportError(this._locale.getString("Error_not_executable"));
			this.openOptions(this._locale.getString("Error_not_executable"));
			return false;
		}

		// Run the editor.
		var process = Components.classes["@mozilla.org/process/util;1"]
			.createInstance(Components.interfaces.nsIProcess);
		process.init(file);
		var args = new Array();
		if (argumentList) {
			// Convert the argument string into an array.
			args = args.concat(argumentList.split(/\s/));

			// Voodoo to handle quotes.
			// Join quoted arguments into one argument.
			var inQuote = -1;
			for (var i = 0; i < args.length; ++i) {
				if (args[i][0] == '"' && inQuote == -1) {
					inQuote = i;
				}
				if (args[i][args[i].length - 1] == '"' && inQuote != -1) {
					// Up to but not including.
					var replacement = args.slice(inQuote, i + 1);
					var joined = replacement.join(' ');
					args.splice(inQuote, replacement.length, joined);
					i = i - (i - inQuote);
					inQuote = -1;
				}
			}
	
			var line = this.getCaretLine();		
			// Replace '%' by the filename and path.
			// We do not have to do so if the user did not specify arguments.
			var replaced = false;
			for (var i = 0; i < args.length; ++i) {
				args[i] = args[i].replace(/"/g, '');
				if (args[i].search(/%l/) != -1) {
					args[i] = args[i].replace(/%l/, line);
				} else if (args[i].search(/%/) != -1) {
					args[i] = args[i].replace(/%/, filename);
					replaced = true;
					// Better check every argument?
					// break;
				}
			}
			// In case there is no '%' we just append the filename.
			if (!replaced) {
				args.push(filename);
			}
		} else {
			args.push(filename);
		}

		process.run(true, args, args.length);

		return true;
	},

	runApp : function() {
		var fileName = "externaleditor";
		var fileExtension = this._prefBranch.getCharPref("fileext");
		// Do not bother adding a default extension, the user probably
		// does not want one then.
		if (fileExtension) {
			fileName += fileExtension;
		}

		var tempFile = new TempFile();
		tempFile.setFileName(fileName);
		if (!tempFile.openFile()) {
			return false;
		}

		// Do not write anything if there is no text.
		if (gContextMenu.target.value !== '') {
			var theText = gContextMenu.target.value;
			// Convert Unix newlines to standard network newlines.
			if (navigator.platform == "Win32") {
				theText = theText.replace(/\n/g, "\r\n");
			}

			tempFile.writeString(theText);
		}

		if (!this.editFile(tempFile.getAbsolutePath())) {
			return false;
		}

		gContextMenu.target.value = tempFile.readString();

		return tempFile.closeFile();
	}
}

function ExternalEditorTB()
{
	// FIXME: Do I need this function?
}

// This actually creates a new instance...
// The new instance will do nothing though.
ExternalEditorTB.prototype = new ExternalEditor();
ExternalEditorTB.prototype.constructor = ExternalEditorTB;

ExternalEditorTB.prototype.runApp = function() {
	var fileName = "externaleditor";
	var fileExtension = this._prefBranch.getCharPref("fileext");
	// Do not bother adding a default extension, the user probably
	// does not want one then.
	if (fileExtension) {
		fileName += fileExtension;
	}

	var tempFile = new TempFile();
	tempFile.setFileName(fileName);
	if (!tempFile.openFile()) {
		return false;
	}

	// XXX: All of this is pretty hacky, fix it!
	var textContainer = document.getElementById('content-frame');
	// Hm?
	// textContainer.editortype = 'text';
	var ed = textContainer.getEditor(textContainer.contentWindow);
	var theText = ed.outputToString('text/plain', 0);

	// Do not write anything if there is no text.
	if (theText !== '') {
		// Convert Unix newlines to standard network newlines.
		if (navigator.platform == "Win32") {
			theText = theText.replace(/\n/g, "\r\n");
		}

		tempFile.writeString(theText);
	}

	if (!this.editFile(tempFile.getAbsolutePath())) {
		return false;
	}

	// FIXME: Clear the editor!
	// textContainer.clearEditor();
	var ed = textContainer.getEditor(textContainer.contentWindow);
	ed.insertText(tempFile.readString());

	return tempFile.closeFile();

	return true;
}

/*
   This class handles a temporary file.
 */
function TempFile()
{
	this.fileName = "temporary";
}

TempFile.prototype = {
	fileName : null,
	tempFile : null,

	openFile : function() {
		var dirService = Components.classes['@mozilla.org/file/directory_service;1'].getService(Components.interfaces.nsIProperties);
		this.tempFile = dirService.get("TmpD", Components.interfaces.nsILocalFile);
		this.tempFile.appendRelativePath(this.fileName);
		this.tempFile.createUnique(this.tempFile.NORMAL_FILE_TYPE, 0600);
		// This might help on some platforms, and with links and such.
		this.tempFile.normalize();
		return true;
	},

	closeFile : function() {
		if (this.tempFile.exists()) {
			this.tempFile.remove(false);
		}
		return true;
	},

	writeString : function(string) {
		var charset = 'UTF-8';
		var fileOutputStream = Components.classes['@mozilla.org/network/file-output-stream;1']
			.createInstance(Components.interfaces.nsIFileOutputStream);
		fileOutputStream.init(this.tempFile, 2, 0x200, false);
		
		var outputStream = Components.classes['@mozilla.org/intl/converter-output-stream;1']
			.createInstance(Components.interfaces.nsIConverterOutputStream);
		outputStream.init(fileOutputStream, charset, 4096, "?".charCodeAt(0));

		outputStream.writeString(string);
		
		outputStream.close();
		fileOutputStream.close();
		return true;
	},

	readString : function() {
		var charset = 'UTF-8';
		var fileInputStream = Components.classes['@mozilla.org/network/file-input-stream;1']
			.createInstance(Components.interfaces.nsIFileInputStream);
		//What do these arguments mean? 0x400 should be permissions (read-only).
		fileInputStream.init(this.tempFile, 1, 0x400, false);

		const replacementChar = Components.interfaces.nsIConverterInputStream
			.DEFAULT_REPLACEMENT_CHARACTER;

		var inputStream = Components.classes['@mozilla.org/intl/converter-input-stream;1']
			.createInstance(Components.interfaces.nsIConverterInputStream);

		inputStream.init(fileInputStream, charset, 4096, replacementChar);

		var str = {};
		var res = '';
		while (inputStream.readString(4096, str) != 0) {
			res += str.value;
		}

		inputStream.close();
		fileInputStream.close();
		return res;
	},

	setFileName : function(string) {
		this.fileName = string;
	},

	getFileName : function() {
		return this.fileName;
	},
	
	getAbsolutePath : function() {
		return this.tempFile.path;
	}
}
